# frozen_string_literal: true
require 'dry/container/nested/version'
require 'dry/container'
require 'forwardable'
require 'inflecto'

module Dry
  class Container
    # @example
    #   module Example
    #     extend Dry::Container::Mixin
    #
    #     register :key, Object.new
    #     register :callable, proc { 1 }, call: false
    #     module Nested
    #       extend Dry::Container::Nested
    #       register :another, Object.new
    #     end
    #   end
    module Nested
      Undefined = Object.new do
        class << self
          def inspect
            'Undefined'
          end

          alias_method :to_str, :inspect
          alias_method :to_s, :inspect
        end
      end.freeze

      # rubocop:disable Metrics/BlockLength

      refine Module do
        # @param [String] name
        # @return [String]
        def containing_path(name = self.name)
          path(name)[0..-2].join('/')
        end

        # @return [Module]
        def containing_module
          Inflecto.constantize Inflecto.camelize containing_path
        end

        # @return [String]
        # @example
        #   module Example
        #     key #=> 'example'
        #     key #=> 'example'
        #     module Nested
        #       key #=> 'example.nested'
        #     end
        #   end
        def key(key = Undefined)
          @__key__ = key unless key == Undefined
          @__key__ ||= path(name).last
        end

        # @param key [String?]
        # @return [String]
        def prefixed(key = nil)
          [prefix, key].compact.join('.')
        end

        def prefix
          container? ? nil : "#{containing_module.prefix}#{key}"
        end

        def namespace_dsl
          container[prefixed]
        end

        def path(name = self.name)
          Inflecto.underscore(name).split('/')
        end

        def container?
          singleton_class.included_modules.include? Dry::Container::Mixin
        end

        def container
          if containing_module.container?
            containing_module
          else
            containing_module.container
          end
        end

        def matcher
          /\A#{prefix}\./
        end

        def deprefixed(key)
          key.gsub(matcher, '')
        end
      end

      # rubocop:enable Metrics/BlockLength

      using self

      # Error
      class NoNameError < ArgumentError
        def initialize(other)
          super("#{other.inspect} cannot be used as a nested container, "\
                'because it does not have a name')
        end
      end

      def self.extended(other)
        super
        raise NoNameError, other unless other.name
      end

      # Class-level DSL for nested container definition
      module ClassInterface
        using Nested
        include Forwardable

        def apply_namespace!
          key_methods.map do |method|
            redefine_method method
          end
        end

        private

        # Methods accepting key as one of their parameters
        # @return [<UnboundMethod>]
        def key_methods
          Mixin.instance_methods(false).inject([]) do |result, method|
            key_parameter(method) ? result + [method] : result
          end
        end

        # @param [UnboundMethod] method_or_name
        def key_parameter(method_or_name)
          method_or_name = method_definition(method_or_name)
          method_or_name.parameters.detect { |_kind, name| :key == name }
        end

        # @param [UnboundMethod, Symbol] method_or_name
        # @param [Module] from
        # @return [UnboundMethod]
        def method_definition(method_or_name, from: Mixin)
          if method_or_name.is_a?(Symbol)
            method_or_name = from.instance_method method_or_name
          end
          unless method_or_name.is_a? UnboundMethod
            raise ArgumentError,
                  "#{method_or_name} is not an unbound method or symbol "\
                  "found in #{from}",
                  caller.join("\n")
          end
          method_or_name
        end

        # @param [UnboundMethod] method
        def redefine_method(method)
          method = method_definition(method)
          key_parameter = key_parameter(method)
          index = method.parameters.index(key_parameter)
          define_method method.name do |*args, **kwargs|
            args = args.dup
            args[index] = prefixed(args[index])
            container.send(method.name, *args, **kwargs)
          end
        end
      end

      extend ClassInterface
      apply_namespace!

      # An array of registered names for the container
      #
      # @return [Array<String>]
      def keys
        container.keys.each_with_object([]) do |key, keys|
          keys << deprefixed(key) if matcher.match?(key)
        end
      end

      # Calls block once for each key in container, passing the key as a
      # parameter.
      #
      # If no block is given, an enumerator is returned instead.
      #
      # @return [Dry::Container::Mixin] self
      def each_key(&block)
        keys.each_key(&block)
        self
      end

      # Calls block once for each key/value pair in the container, passing the
      # key and the registered item parameters.
      #
      # If no block is given, an enumerator is returned instead.
      #
      # @return [Enumerator]
      def each(&block)
        container.config
                 .resolver
                 .each(_container)
                 .each_with_object([]) do |(key, value), result|
          result << [deprefixed(key), value] if matcher.match?(key)
        end.each(&block)
      end
    end
  end
end
